from sbnltk.Tokenizer import wordTokenizer
from sbnltk.Preprocessor import preprocessor
from sbnltk.Stemmer import stemmerOP
from sbnltk.Downloader import  downloader
from sbnltk import sbnltk_default
import pickle

class static_postag:
    __dl=downloader()
    __dict={}
    __stemmer=None
    __bp=preprocessor()
    __tokenizer=wordTokenizer()
    def __init__(self):
        self.__dl.download('postag_static',sbnltk_default.sbnltk_root_path+'dataset/')
        self.__stemmer=stemmerOP()
        path=sbnltk_default.sbnltk_root_path+'dataset/postag_static.txt'
        for word in open(path,'r'):
            word=word.replace('\n','')
            tokens=self.__tokenizer.basic_tokenizer(word)
            wd=tokens[0]
            val=tokens[-1]
            self.__dict[wd]=val
    def tag(self,sent):
        tokens=self.__tokenizer.basic_tokenizer(sent)
        ans=[]
        for word in tokens:
            if self.__bp.is_number(word):
                ans.append((word,'NUM'))
                continue
            if self.__dict.get(word):
               ans.append((word,self.__dict[word]))
               continue
            if self.__dict.get(self.__bp.word_normalize(word)) :
                ans.append((word, self.__dict[self.__bp.word_normalize(word)]))
                continue
            stem_word=self.__stemmer.stemWord(word)
            if self.__dict.get(stem_word):
               ans.append((word,self.__dict[stem_word]))
               continue
            ans.append((word,'unk'))
        return ans


class sklearn_postag:
    __dl=downloader()
    __bp=preprocessor()
    __sk_model=None
    def __init__(self):
        self.__dl.download('sklearn_postag',sbnltk_default.sbnltk_root_path+'model/')
        self.__sk_model = pickle.load(open(sbnltk_default.sbnltk_root_path+'model/sklearn_postag.pkl', 'rb'))
    def word2features(self,sent, i):
        return {
            'word': sent[i], 'is_first': i == 0, 'is_last': i == len(sent) - 1,
            'is_capitalized': sent[i][0].upper() == sent[i][0], 'is_all_caps': sent[i].upper() == sent[i],
            'is_all_lower': sent[i].lower() == sent[i],
            'prefix-1': sent[i][0], 'prefix-2': sent[i][:2], 'prefix-3': sent[i][:3],
            'suffix-1': sent[i][-1], 'suffix-2': sent[i][-2:], 'suffix-3': sent[i][-3:],
            'prev_word': '' if i == 0 else sent[i - 1], 'next_word': '' if i == len(sent) - 1 else sent[i + 1],
            'is_numeric': sent[i].isdigit()
        }
    def tag(self,text):
        if len(text)==0:
            return []
        words=text.split()
        sentence_features = [self.word2features(words, i) for i in range(len(words))]
        return list(zip(words, self.__sk_model.predict([sentence_features])[0]))



from Sentence_embedding import sentence_embedding_from_word2vec,Bangla_sentence_embedding_gd

class information_extraction_word2vec:
    __s_w2v=None
    def __init__(self):
        self.__s_w2v=sentence_embedding_from_word2vec()
    def compute_information_from_title(self,title,sentences,k):
        if (len(sentences) == 0):
            return []
        dict = {}
        vec = []
        for k in range(min(k, len(sentences))):
            low = 10000.0
            ind = 0
            i = 0
            for sent in sentences:
                d = self.__s_w2v.dist(sent, title)
                if d < low and (i not in dict):
                    low = d
                    ind = i
                i += 1
            dict[ind] = 1
            vec.append([low, ind, sentences[ind]])
        return sorted(vec)
    def __compute_centroid(self,sentences):
        low = 10000.0
        ind = 0

        for i in range(len(sentences)):
            sum = 0.0
            for j in range(len(sentences)):
                sum += self.__s_w2v.dist(sentences[i], sentences[j])
            if sum < low:
                low = sum
                ind = i
        return ind
    def compute_information_from_single_centroid(self,sentences,k):
        if(len(sentences)==0):
            return []
        title_ind=self.__compute_centroid(sentences)
        return self.compute_information_from_title(sentences[title_ind],sentences,k)


    def compute_information_from_neighbours(self,sentences,k):
        if (len(sentences) == 0):
            return []
        title_ind = self.__compute_centroid(sentences)
        dict = {}
        dict[title_ind] = 1
        title = sentences[title_ind]
        low = 10000.0
        ind = 0
        i = 0
        vec = []
        for k in range(min(k, len(sentences))):
            i = 0
            for sent in sentences:
                d = self.__s_w2v.dist(sent, title)
                if d < low and (i not in dict):
                    low = d
                    ind = i
                i += 1
            dict[ind] = 1
            vec.append([low, ind, sentences[ind]])
        return sorted(vec)

class information_extraction_transformer_embedding:
    __s_tse=None
    def __init__(self):
        self.__s_tse = Bangla_sentence_embedding_gd()

    def n_gram_sentence(self,title,sentences,n):
        title_embeddings = self.__s_tse.encode_single_sentence(title)
        sentence_embeddings = self.__s_tse.encode_sentence_list(sentences)
        top_sentences = []
        done_sentence = []
        embedding = title_embeddings
        for i in range(len(sentences) - 1):
            j = 0
            mx = 0.0
            temp = ""
            ind = 0
            for sentence in sentences:
                if j in done_sentence:
                    j += 1
                    continue
                sim = self.__s_tse.similarity_of_two_embedding(embedding, sentence_embeddings[sentence])
                if sim > mx:
                    mx = sim
                    temp = sentence
                    embedding = sentence_embeddings[sentence]
                    ind = j
                j += 1
            done_sentence.append(ind)
            top_sentences.append([mx, ind, temp])
            n=min(n,len(sentences))
        return top_sentences[:n]
    def compute_information_from_title(self,title,sentences,k):
        title_embeddings = self.__s_tse.encode_single_sentence(title)
        sentence_embeddings = self.__s_tse.encode_sentence_list(sentences)
        top_sentences = []
        ind = 0
        k=min(k,len(sentences))
        for sentence in sentences:
            sim = self.__s_tse.similarity_of_two_embedding(title_embeddings, sentence_embeddings[sentence])
            top_sentences.append([sim, ind, sentence])
            ind += 1
        top_sentences = sorted(top_sentences, reverse=True)
        return top_sentences[:k]
    def __compute_centroid(self,sentences):
        high = 0.0
        ind = 0
        x=0
        sentence_embeddings = self.__s_tse.encode_sentence_list(sentences)
        for i in sentences:
            sum = 0.0
            for j in sentences:
                sum += self.__s_tse.similarity_of_two_embedding(sentence_embeddings[i], sentence_embeddings[j])
            if sum > high:
                high = sum
                ind = x
            x+=1
        return ind
    def compute_information_from_centroid(self,sentences,k):
        title=sentences[self.__compute_centroid(sentences)]
        k = min(k, len(sentences))
        return self.compute_information_from_title(title,sentences,k)


'''
Types of SentimentAnalyzer
1. logisticRegression
2. LinearSVC
3. Multinomial_naive_bayes
5. Random Forest
4. Bert Sentiment
5. SVC
'''

from sbnltk.Downloader import downloader
import pickle
import pandas as pd
import numpy as np
from transformers import *
import tensorflow as tf
import tensorflow_addons as tfa
from sbnltk import sbnltk_default


class sentimentAnalyzer:
    __dl=downloader()
    __sentiment_models=[('LR','Logistic Regression'),('LSVC','Linear SVC'),
                      ('MNB','Multinomial naive bayes'),('RF','Random Forest'),('BERT','Bert Sentiment Analysis')
                      ]
    __root_path=sbnltk_default.sbnltk_root_path
    def all_sentiment_models(self):
        st='All Sentiment analysis models name with code\n'
        for sent in self.__sentiment_models:
            st+=sent[1]+' ::: '+sent[0]+'\n'
        return st
    def __LR(self,sentences):
        self.__dl.download('sentiment_LR',sbnltk_default.sbnltk_root_path+'model/')
        self.__dl.download('sentiment_vector',sbnltk_default.sbnltk_root_path+'model/')
        logreg = pickle.load(open(sbnltk_default.sbnltk_root_path+'model/sentiment_LR.pkl', 'rb'))
        vectorizer = pickle.load(open(sbnltk_default.sbnltk_root_path+'model/sentiment_vector.pkl', 'rb'))
        unknown_vectors = vectorizer.transform(sentences)
        unknown_words_df = pd.DataFrame(unknown_vectors.toarray(), columns=vectorizer.get_feature_names())
        pred=[]
        prop=[]
        for i in range(len(sentences)):
            pred.append(logreg.predict(unknown_words_df)[i])
            prop.append(logreg.predict_proba(unknown_words_df)[:, 1][i])
        return pred,prop
    def __LSVC(self,sentences):
        self.__dl.download('sentiment_LSVC', sbnltk_default.sbnltk_root_path+'model/')
        self.__dl.download('sentiment_vector', sbnltk_default.sbnltk_root_path+'model/')
        svc = pickle.load(open(sbnltk_default.sbnltk_root_path+'model/sentiment_LSVC.pkl', 'rb'))
        vectorizer = pickle.load(open(sbnltk_default.sbnltk_root_path+'model/sentiment_vector.pkl', 'rb'))
        unknown_vectors = vectorizer.transform(sentences)
        unknown_words_df = pd.DataFrame(unknown_vectors.toarray(), columns=vectorizer.get_feature_names())
        pred = []
        for i in range(len(sentences)):
            pred.append(svc.predict(unknown_words_df)[i])
        return pred
    def __MNB(self,sentences):
        self.__dl.download('sentiment_MNB',sbnltk_default.sbnltk_root_path+'model/')
        self.__dl.download('sentiment_vector',sbnltk_default.sbnltk_root_path+'model/')
        mnb = pickle.load(open(sbnltk_default.sbnltk_root_path+'model/sentiment_MNB.pkl', 'rb'))
        vectorizer = pickle.load(open(sbnltk_default.sbnltk_root_path+'model/sentiment_vector.pkl', 'rb'))
        unknown_vectors = vectorizer.transform(sentences)
        unknown_words_df = pd.DataFrame(unknown_vectors.toarray(), columns=vectorizer.get_feature_names())
        pred = []
        prop = []
        for i in range(len(sentences)):
            pred.append(mnb.predict(unknown_words_df)[i])
            prop.append(mnb.predict_proba(unknown_words_df)[:, 1][i])
        return pred, prop
    def __RF(self,sentences):
        self.__dl.download('sentiment_RF',sbnltk_default.sbnltk_root_path+'model/')
        self.__dl.download('sentiment_vector',sbnltk_default.sbnltk_root_path+'model/')
        rf = pickle.load(open(sbnltk_default.sbnltk_root_path+'model/sentiment_RF.pkl', 'rb'))
        vectorizer = pickle.load(open(sbnltk_default.sbnltk_root_path+'model/sentiment_vector.pkl', 'rb'))
        unknown_vectors = vectorizer.transform(sentences)
        unknown_words_df = pd.DataFrame(unknown_vectors.toarray(), columns=vectorizer.get_feature_names())
        pred = []
        prop = []
        for i in range(len(sentences)):
            pred.append(rf.predict(unknown_words_df)[i])
            prop.append(rf.predict_proba(unknown_words_df)[:, 1][i])
        return pred, prop
    def __sentence_convert_data(self,data):
        tokenizer = BertTokenizer.from_pretrained(sbnltk_default.sbnltk_root_path+'model/sentiment_multilingual_vocab.txt')
        SEQ_LEN = 147
        tokens, masks, segments = [], [], []
        token = tokenizer.encode(data, max_length=SEQ_LEN, truncation=True, padding='max_length')
        num_zeros = token.count(0)
        mask = [1] * (SEQ_LEN - num_zeros) + [0] * num_zeros
        segment = [0] * SEQ_LEN
        tokens.append(token)
        segments.append(segment)
        masks.append(mask)
        tokens = np.array(tokens)
        masks = np.array(masks)
        segments = np.array(segments)
        return [tokens, masks, segments]

    def __b_predict(self,bert,sentences):
        pred=[]
        prop=[]
        for sent in sentences:
            data_x = self.__sentence_convert_data(sent)
            predict = bert.predict(data_x)
            predict_value = np.ravel(predict)
            predict_answer = np.round(predict_value, 0).item()
            if predict_answer == 0:
                pred.append(0)
                prop.append((1.0 - predict_value[0]))
            else:
                pred.append(1)
                prop.append((predict_value[0]))
        return pred,prop

    def __create_sentiment_bert(self):
        SEQ_LEN = 147
        model = TFBertModel.from_pretrained('bert-base-multilingual-cased')
        token_inputs = tf.keras.layers.Input((SEQ_LEN,), dtype=tf.int32, name='input_word_ids')
        mask_inputs = tf.keras.layers.Input((SEQ_LEN,), dtype=tf.int32, name='input_masks')
        segment_inputs = tf.keras.layers.Input((SEQ_LEN,), dtype=tf.int32, name='input_segment')
        bert_outputs = model([token_inputs, mask_inputs, segment_inputs])
        bert_outputs = bert_outputs[1]
        sentiment_first = tf.keras.layers.Dense(1, activation='sigmoid',
                                                kernel_initializer=tf.keras.initializers.TruncatedNormal(stddev=0.02))(
            bert_outputs)
        sentiment_model = tf.keras.Model([token_inputs, mask_inputs, segment_inputs], sentiment_first)
        opt = tfa.optimizers.RectifiedAdam(lr=2.0e-5, weight_decay=0.0025)
        sentiment_model.compile(optimizer=opt, loss=tf.keras.losses.BinaryCrossentropy(), metrics=['acc'])
        return sentiment_model
    def __BERT(self,sentence):
        self.__dl.download('sentiment_BERT',sbnltk_default.sbnltk_root_path+ 'model/')
        self.__dl.download('sentiment_multilingual_vocab',sbnltk_default.sbnltk_root_path+'model/')
        bert = self.__create_sentiment_bert()
        bert.load_weights(sbnltk_default.sbnltk_root_path+'model/sentiment_BERT.h5')
        return self.__b_predict(bert,sentence)
    def predict(self,model_code,sentences):
        if len(sentences)==0:
            raise ValueError('Empty list of Sentences is detected in Sentiment analysis!!')
        if model_code=='LR':
               pred,prop=self.__LR(sentences)
               return pred,prop
        elif model_code=='LSVC':
            pred=self.__LSVC(sentences)
            return pred
        elif model_code=='MNB':
            pred,prop=self.__MNB(sentences)
            return pred,prop
        elif model_code=='RF':
            pred,prop=self.__RF(sentences)
            return pred,prop
        elif model_code=='BERT':
            pred,prop=self.__BERT(sentences)
            return pred,prop
        else:
            raise ValueError('Model code Does not exist!!\n'+ self.all_sentiment_models())






'''
WordTokenizer
    basic_tokenizer: only split words by space
    customized_tokenizer: split words, can remove punctuation (initially True), can make normalization(initially False), can remove dust(intitially False)

SentenceTokenizer:
    basic_tokenizer: only split by Bangla_fullstop
    customized_tokenizer: split sentences, can remove punctuation (initially True), can make normalization(initially False), can remove dust(intitially False)
    sentence_cluster: clustering sentences with max_length, can remove punctuation (initially True), can make normalization(initially False), can remove dust(intitially False)

'''



from sbnltk.Preprocessor import preprocessor
from sbnltk import sbnltk_default
bp=preprocessor()
class wordTokenizer:
    def basic_tokenizer(self,text):
        tokens = text.split()
        return tokens

    def customized_tokenizer(self,text,punc=True,norm=False,dust=False):
        if punc==True:
            text=bp.punctuation_remove(text)
        if norm==True:
            text=bp.word_normalize(text)
        tokens=text.split()
        try:
            if dust==True:
                temp_tokens=[]
                for i in tokens:
                    if len(bp.dust_removal(i))==0:
                        continue
                    i=bp.dust_removal(i)
                    temp_tokens.append(i)
                return temp_tokens
            else:
                return tokens
        except:
            print(f"{sbnltk_default.bcolors.FAIL} ERROR 301: Error in Customized word Tokenizer!! {sbnltk_default.bcolors.ENDC}")
            return tokens

class sentenceTokenizer:

    __pre=preprocessor()

    def basic_tokenizer(self,text):
        text=text.replace('\n',' ')
        tokens = []
        s = ""
        bangla_fullstop = '0964'
        for c in text:
            g = c.encode("unicode_escape")
            g = g.upper()
            g = g[2:]
            g = g.decode('utf-8')
            if g == bangla_fullstop:
                tokens.append(s)
                s = ""
                continue
            s += c
        if len(s) > 0:
            tokens.append(s)
        return tokens

    def customized_tokenizer(self,text,punc=True,norm=False,dust=False):
        tokens=[]
        text = text.replace('\n', ' ')
        s=""
        bangla_fullstop = '0964'
        for c in text:
            g = c.encode("unicode_escape")
            g = g.upper()
            g = g[2:]
            g = g.decode('utf-8')
            if g==bangla_fullstop:
                tokens.append(s)
                s=""
                continue
            s+=c
        if len(s)>0:
            tokens.append(s)
        try:
            temp_tokens=[]
            for i in tokens:
                if punc==True:
                    i=bp.punctuation_remove(i)
                if norm==True:
                    i=bp.word_normalize(i)
                if len(bp.dust_removal_sent(i))!=0 and dust==True:
                    i=bp.dust_removal_sent(i)
                temp_tokens.append(i)
            return temp_tokens
        except:
            print(f"{sbnltk_default.bcolors.FAIL} ERROR 302: Error in Customized Sentence Tokenizer!! {sbnltk_default.bcolors.ENDC}")
            return tokens

    def sentence_vector_to_text(self,sentences,full_stop=True):
        if full_stop==True:
            text=sbnltk_default.bangla_full_stop.join(sentences)
            text+=sbnltk_default.bangla_full_stop
        else:
            text=' '.join(sentences)
            text=self.__pre.extra_space_remove(text)
        return text

    def sentence_cluster(self,text,max_length=100,punc=True,norm=False,dust=False):
        tokens = []
        text = text.replace('\n', ' ')
        s = ""
        bangla_fullstop = '0964'
        for c in text:
            g = c.encode("unicode_escape")
            g = g.upper()
            g = g[2:]
            g = g.decode('utf-8')
            if g == bangla_fullstop:
                tokens.append(s)
                s = ""
                continue
            s += c
        if len(s) > 0:
            tokens.append(s)
        try:
            '''
            tmp_tokens: temporary tokens for returning
            word_tokens: taking word from each string
            tmp_sent: temporary sentence which length at most : max length
            '''
            tmp_tokens=[]
            for sent in tokens:
                if len(sent)>max_length:
                    l=len(sent)
                    word_tokens=sent.split()
                    tmp_sent=''
                    for w in word_tokens:
                        if len(str(tmp_sent+w))>max_length:
                            tmp_sent=tmp_sent[:-1]
                            tmp_tokens.append(tmp_sent)
                            tmp_sent=''
                        tmp_sent=tmp_sent+w+' '
                    tmp_tokens.append(tmp_sent)
                else:
                    tmp_tokens.append(sent)
            tmp_tokens2=[]
            for sent in tmp_tokens:
                if punc==True:
                    sent=bp.punctuation_remove(sent)
                if norm==True:
                    sent=bp.word_normalize(sent)
                if len(bp.dust_removal_sent(sent)) != 0 and dust == True:
                    sent = bp.dust_removal_sent(sent)
                tmp_tokens2.append(sent)
            return tmp_tokens2
        except:
            print(f"{sbnltk_default.bcolors.FAIL} ERROR 303: Error in Sentence clustering!! {sbnltk_default.bcolors.ENDC}")
            return tokens








from sbnltk.Tokenizer import wordTokenizer
from sbnltk.word_embedding import gensim_word2vec_embedding
import math
import os
import zipfile
from sbnltk.Downloader import downloader
from sentence_transformers import SentenceTransformer,util
from sbnltk import sbnltk_default

class sentence_embedding_from_word2vec:
    __w2v=None
    __tokenizer=None
    def __init__(self):
        self.__tokenizer=wordTokenizer()
        self.__w2v=gensim_word2vec_embedding()
    def dist(self,sent1,sent2):
        tokens1=self.__tokenizer.basic_tokenizer(sent1)
        tokens2=self.__tokenizer.basic_tokenizer(sent2)
        d=0
        for i in tokens1:
            mn=math.inf
            for j in tokens2:
                mn=min(mn,1.0-self.__w2v.cosine_distance(i,j))
            d+=mn
        return d
    # root sentence finding
    def root_sentence_finding(self,vec):
        ans=[]
        mx=0
        for sent1 in vec:
            temp=[]
            for sent2 in vec:
                temp.append(self.dist(sent1,sent2))
                mx=max(mx,self.dist(sent1,sent2))
            ans.append(temp)
        temp_ans=[]
        mx+=1.0
        for i in ans:
            temp=[]
            for j in i:
                temp.append((mx/(j+1.0)))
            temp_ans.append(temp)
        return temp_ans

class Bangla_sentence_embedding_hd:
    __dl=downloader()
    __model=None
    def __init__(self):
        if os.path.exists(sbnltk_default.sbnltk_root_path+'model/sentence_embedding_transformer')==False:
            self.__dl.download('sentence_embedding_transformer_hd', sbnltk_default.sbnltk_root_path + 'model/')
            with zipfile.ZipFile(sbnltk_default.sbnltk_root_path+'model/sentence_embedding_transformer_hd.zip', 'r') as file:
                file.extractall(sbnltk_default.sbnltk_root_path+'model/')
            os.remove(sbnltk_default.sbnltk_root_path+'model/sentence_embedding_transformer_hd.zip')
        self.__model=SentenceTransformer(sbnltk_default.sbnltk_root_path+'model/sentence_embedding_transformer')
    def encode_sentence_list(self,sentences):
        embeddings={}
        sentence_embeddings=self.__model.encode(sentences)
        for sentence, embedding in zip(sentences, sentence_embeddings):
            embeddings[sentence]=embedding
        return embeddings
    def similarity_of_two_sentence(self,sentence1,sentence2):
        embed=self.encode_sentence_list([sentence1,sentence2])
        return util.pytorch_cos_sim(embed[sentence1], embed[sentence2])
    def similarity_of_two_embedding(self,embedding1,embedding2):
        return util.pytorch_cos_sim(embedding1, embedding2)

class Bangla_sentence_embedding_gd:
    __dl = downloader()
    __model = None
    def __init__(self):
        if os.path.exists(sbnltk_default.sbnltk_root_path + 'model/Towhid-Sust-transformer') == False:
            self.__dl.download('sentence_embedding_transformer_gd', sbnltk_default.sbnltk_root_path + 'model/')
            with zipfile.ZipFile(sbnltk_default.sbnltk_root_path + 'model/sentence_embedding_transformer_gd.zip',
                                 'r') as file:
                file.extractall(sbnltk_default.sbnltk_root_path + 'model/')
            os.remove(sbnltk_default.sbnltk_root_path + 'model/sentence_embedding_transformer_gd.zip')
        self.__model = SentenceTransformer(sbnltk_default.sbnltk_root_path + 'model/Towhid-Sust-transformer')

    def encode_sentence_list(self, sentences):
        embeddings = {}
        sentence_embeddings = self.__model.encode(sentences)
        for sentence, embedding in zip(sentences, sentence_embeddings):
            embeddings[sentence] = embedding
        return embeddings
    def encode_single_sentence(self,sentence):
        return self.__model.encode(sentence)
    def similarity_of_two_sentence(self, sentence1, sentence2):
        embed = self.encode_sentence_list([sentence1, sentence2])
        return util.pytorch_cos_sim(embed[sentence1], embed[sentence2])

    def similarity_of_two_embedding(self, embedding1, embedding2):
        return util.pytorch_cos_sim(embedding1, embedding2)


# tf(0.1)
# position(0.1)
# Numeric word(0.3)
# Sentence length(0.1)
# matched with title(0.1)
# exact title occurs(0.1)
from sbnltk.NER import sklearn_NER
from sbnltk.Stemmer import stemmerOP
from sbnltk.Tokenizer import wordTokenizer,sentenceTokenizer
from sbnltk.Preprocessor import preprocessor
from sbnltk.Postag import sklearn_postag

class feature_based_information_extraction:
    __pre=None
    __tokenizer=None
    __sentT=None
    __stemmer=None
    __sentences=[]
    __sent_pos={}
    __calc={}
    __title=""
    __nerT=None
    __posT=None
    def __init__(self):
        self.__stemmer=stemmerOP()
        self.__pre=preprocessor()
        self.__tokenizer=wordTokenizer()
        self.__sentT=sentenceTokenizer()
        self.__posT=sklearn_postag()
    def __term_frequency(self):
        word = {}
        total_word = 0
        for i in self.__sentences:
            tmp = self.__pre.punctuation_remove(i)
            tmp = self.__pre.stopword_remove(tmp)
            if len(tmp) == 0:
                continue
            tokens = self.__tokenizer.basic_tokenizer(tmp)
            for w in tokens:
                total_word += 1
                w = self.__stemmer.stemWord(w)
                if w in word:
                    word[w] += 1
                else:
                    word[w] = 1
        for i in self.__sentences:
            tmp = self.__pre.punctuation_remove(i)
            tmp = self.__pre.stopword_remove(tmp)
            if len(tmp) == 0:
                continue
            tokens = self.__tokenizer.basic_tokenizer(tmp)
            cnt = 0
            for w in tokens:
                cnt += word[self.__stemmer.stemWord(w)]
            self.__calc[i] += (cnt / total_word) * 0.1
        return
    def __is_number(self,word):
        v = ['০', '১', '২', '৩', '৪', '৫', '৬', '৭', '৮', '৯', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
        for c in word:
            if c in v:
                return True
        return False
    def __position(self):
        ln=len(self.__sentences)
        mid=int(0.7*ln)
        pos=1
        now=0.1
        dv1=0.1/mid
        for i in self.__sentences:
            if pos==mid:
                break
            self.__calc[i]+=now
            now-=dv1
            pos+=1

        pos=1
        dv2=.1/(ln-mid)
        now=dv2
        for i in self.__sentences:
            if pos<mid:
                pos+=1
                continue
            self.__calc[i]+=now
            now+=dv2
            pos+=1
        return
    def __Numeric_word(self):
        for i in self.__sentences:
            tokens = self.__tokenizer.basic_tokenizer(i)
            flg=0
            for w in tokens:
                if self.__is_number(w):
                    flg=1
                    break
            if flg:
                self.__calc[i]+=0.1
        return
    def __sentence_length(self):
        for i in self.__sentences:
            if len(i)>50:
                continue
            self.__calc[i]+=((50-len(i))/50)*0.1
        return
    def __matched_with_title(self):
        title=self.__stemmer.stemSent(self.__title)
        title_tokens=self.__tokenizer.basic_tokenizer(title)

        for i in self.__sentences:
            cnt=0
            e_title=0
            now=title_tokens[0]
            pos=0
            tmp={}
            tm_sent=i
            i=self.__stemmer.stemSent(i)
            tokens=self.__tokenizer.basic_tokenizer(i)
            for w in tokens:
                if w in title_tokens:
                    if w not in tmp:
                        cnt+=1
                        tmp[w]=1
                if pos>=len(title_tokens):
                    continue
                if w==now:
                    e_title+=1
                    pos+=1
                    if pos<len(title_tokens):
                        now=title_tokens[pos]
            self.__calc[tm_sent]+=(len(tmp)/len(title_tokens))*0.1
            if pos>=len(title_tokens):
                self.__calc[tm_sent]+=0.1
        return
    def __count_ner_tags_with_tag(self,sentence):
        cnt=0
        for i in sentence:
            if i[1]!='O':
                cnt+=1
        return cnt

    def __count_ner(self):
        self.__nerT=sklearn_NER()
        for sent in self.__sentences:
            tags = self.__nerT.tag(sent)
            cnt = self.__count_ner_tags_with_tag(tags)
            if len(tags) > 0:
                self.__calc[sent] += float(float(cnt * 0.1) / len(tags))
        return

    def __count_postag(self):
        for sent in self.__sentences:
            tags=self.__posT.tag(sent)
            cnt=0
            for i in tags:
                if i[1]=='NN' or i[1]=='NNP':
                    cnt+=1
            if len(tags)>0:
                self.__calc[sent]+=float(float(cnt*0.1)/float(len(tags)))
        return
    def __update(self):
        pos = 0
        for i in self.__sentences:
            self.__calc[i] = 0.0
            self.__sent_pos[i] = pos
            pos += 1
        self.__term_frequency()
        self.__position()
        self.__Numeric_word()
        self.__sentence_length()
        if len(self.__title)!=0:
            self.__matched_with_title()
        self.__count_ner()
        self.__count_postag()

    '''
        param: text, 
               title = None(default)
               max_output_sent= 5 (default)
        return :
            extracted text 
        '''
    def extract_text_to_text(self,text,title=None,max_extract_sent=5):
        if len(text)==0:
            raise ValueError('Null Text for information Extraction')
        if title!=None:
            self.__title=title
        text=self.__pre.extra_space_remove(text)
        sentences=self.__sentT.basic_tokenizer(text)
        self.__sentences=sentences
        self.__update()
        vec=[]
        for i in sentences:
            vec.append((self.__calc[i],self.__sent_pos[i]))
        top_sentences = sorted(vec, reverse=True)
        max_extract_sent=min(max_extract_sent,len(text))
        ans_vec=[]
        for i in range(len(top_sentences)):
            if i>=max_extract_sent:
                break
            ans_vec.append(top_sentences[i][1])
        ans_temp_vec=sorted(ans_vec)
        ans_vec=[]
        for i in ans_temp_vec:
            ans_vec.append(sentences[i])
        extract_text=self.__sentT.sentence_vector_to_text(ans_vec)
        return extract_text

    '''
    param: text, 
           title = None(default)
           max_output_sent= 5 (default)
    return :
        sorted vector (higher to lower)
        1st position: score of that sentence
        2nd position: sentence
        3rd position: position of sentence
    '''
    def extract_text_to_vector(self,text,title=None,max_extract_sent=5):
        if len(text)==0:
            raise ValueError('Null Text for information Extraction')
        text = self.__pre.extra_space_remove(text)
        if title!=None:
            self.__title=title
        sentences=self.__sentT.basic_tokenizer(text)
        self.__sentences=sentences
        self.__update()
        vec=[]
        for i in sentences:
            vec.append((self.__calc[i],i,self.__sent_pos[i]))
        max_extract_sent = min(max_extract_sent, len(text))
        top_sentences = sorted(vec, reverse=True)[:max_extract_sent]
        return top_sentences








'''

** Bangla Text preprocessor **

Punctuation remove: remove all type of punctuation and replace punctuation with a SPACE
dust_removal : remove all kind of character(including English character, punctuation) without Bangla character Aভাই23 = ভাই
stopword_remove: remove stopword from text
word_normalize: similar Bangla vowel defines same character : অসহনীয় ভারী বর্ষণে : অসহনিয় ভারি বর্ষণে
bangla_to_english_Conversion: Bangla to English word equivalent word conversion : রাজধানী : rajdhani
bnCompare: compare function
isBangla: is Bangla character exist in this word?
bn_word_sort_en_sys: Sort bangla word according to English alphabet (a,b,c,......)
bn_word_sort_bn_sys: Sort bangla word according to Bangla alphabet(অ,.....)
is_number: is bangla number?

'''


import re
import functools
from sbnltk import sbnltk_default
from sbnltk.Downloader import downloader

class StaticArray:
    bn2en={'0985':'a','0986':'a','0987':'i','0988':'i',
           '0989':'u','098A':'u','098B':'r','09E0':'r',
           '098C':'l','09E1':'l','098F':'e','0990':'oi',
           '0993':'o','0994':'ou',
           '0995':'k','0996':'kh','0997':'g','0998':'gh',
           '0999':'ng','099A':'c','099B':'ch','099C':'j',
           '099D':'jh','099E':'ny','099F':'tt', '09A0':'tth',
           '09A1':'dd','09DC':'rr','09A2':'ddh','09DD':'rh',
           '09A3':'nn','09A4':'t','09A5':'th','09A6':'d',
           '09A7':'dh','09A8':'n','09AA':'p','09AB':'ph',
           '09AC':'b','09AD':'bh','09AE':'m','09AF':'y',
           '09DF':'yy','09B0':'r','09B2':'l','09B6':'sh',
           '09B7':'ss','09B8':'s','09B9':'h','09CE':'t',
           '09BE':'a','09BF':'i','09C0':'i','09C1':'u',
           '09C2':'u','09C3':'r','09C4':'r','09C7':'e',
           '09C8':'oi', '09CB':'o','09CC':'au',
           '09CD':'','09D7':'','09E6':'0',
           '09E7':'1','09E8':'2','09E9':'3','09EA':'4',
           '09EB':'5','09EC':'6','09ED':'7','09EE':'8',
           '09EF':'9',
           '09BC':'','0982':'n','0983':'','0981':''
           }
    bn2enNum={
        '09E6': '0',
        '09E7': '1', '09E8': '2', '09E9': '3', '09EA': '4',
        '09EB': '5', '09EC': '6', '09ED': '7', '09EE': '8',
        '09EF': '9'
    }
    bn2enPunc={
        '09BE': 'a', '09BF': 'i', '09C0': 'i', '09C1': 'u',
        '09C2': 'u', '09C3': 'r', '09C4': 'r', '09C7': 'e',
        '09C8': 'ai', '09CB': 'o', '09CC': 'au','09B0':'ra'
    }
    bn2enSuffix={
        '09BE': 'a', '09B0':'ra','09CB': 'o','0993':'o','09C7': 'e'
    }
    bn_norm={
        '09C0':'\u09bf','09C2':'\u09c1','09C4':'\u09c3','09A3':'\u09a8','0988':'\u0987','098A':'\u0989'
    }

    bn_serial={
        '0985': 0, '0986': 1, '09BE': 2, '0987': 3, '09BF': 4,'0988': 5, '09C0': 6,
        '0989': 7, '09C1': 8, '098A': 9, '09C2': 10,'098B': 11,'09C3': 12, '09E0': 13,'09C4': 14,
        '098C': 15, '09E1': 16, '098F': 17,  '09C7': 18, '0990': 19,'09C8': 20,
        '0993': 21, '09CB': 22, '0994': 23, '09CC': 24,
        '0995': 25, '0996': 26, '0997': 27, '0998': 28,
        '0999': 29, '099A': 30, '099B': 31, '099C': 32,
        '099D': 33, '099E': 34, '099F': 35, '09A0': 36,
        '09A1': 37, '09DC': 38, '09A2': 39, '09DD': 40,
        '09A3': 41, '09A4': 42, '09A5': 43, '09A6': 44,
        '09A7': 45, '09A8': 46, '09AA': 47, '09AB': 48,
        '09AC': 49, '09AD': 50, '09AE': 51, '09AF': 52,
        '09DF': 53, '09B0': 54, '09B2': 55, '09B6': 56,
        '09B7': 57, '09B8': 58, '09B9': 59, '09CE': 60,
        '09E6': 61,
        '09E7': 62, '09E8': 63, '09E9': 64, '09EA': 65,
        '09EB': 66, '09EC': 67, '09ED': 68, '09EE': 69,
        '09EF': 70,
        '09CD': 71, '09D7': 72,
        '09BC': 73, '0982': 74, '0983': 75, '0981': 76
    }

class preprocessor:

    __dl=downloader()
    __word_list = {}
    __stopwords=[]
    def __init__(self):
        self.__dl.download('bangla_word_list', sbnltk_default.sbnltk_root_path+'dataset/')
        self.__dl.download('stopword_list', sbnltk_default.sbnltk_root_path+'dataset/')
        for line in open(sbnltk_default.sbnltk_root_path+'dataset/bangla_word_list.txt', 'r'):
            line = line.rstrip('\n')
            self.__word_list[line] = 1
        model_path = sbnltk_default.sbnltk_root_path+"dataset/stopword_list.txt"
        for i in open(model_path, "r"):
            i = i.rstrip("\n")
            self.__stopwords.append(i)

    def punctuation_remove(self,text):
        try:
            whitespace = re.compile(u"[\s\u0020\u00a0\u1680\u180e\u202f\u205f\u3000\u2000-\u200a]+", re.UNICODE)
            bangla_fullstop = u"\u0964"
            punctSeq = u"['\"“”‘’]+|[.?!,…]+|[:;]+"
            punc = u"[(),$%^&*+={}\[\]:\"|\'\~`<>/,¦!?½£¶¼©⅐⅑⅒⅓⅔⅕⅖⅗⅘⅙⅚⅛⅜⅝⅞⅟↉¤¿º;-]+"
            text = whitespace.sub(" ", text).strip()
            text = re.sub(punctSeq, " ", text)
            text = re.sub(bangla_fullstop, " ", text)
            text = re.sub(punc, " ", text)
            text = re.sub('[!"#$%&\'()*+,-./:;<=>?@[\]^_`{|}~]', ' ', text)
            text=text.replace("\\", " ")
            return text
        except:
            print(f"{sbnltk_default.bcolors.FAIL} ERROR 201: Error in Removing punctuation!! {sbnltk_default.bcolors.ENDC}")
            return text


    def dust_removal(self,word):

        try:
            s=""
            for c in word:
                g = c.encode("unicode_escape")
                g = g.upper()
                g = g[2:]
                g = g.decode('utf-8')
                if g in StaticArray.bn2en:
                    s+=c
            if len(s)==0:
                return word
            return s
        except:
            print(f"{sbnltk_default.bcolors.FAIL} ERROR 202: Error in Removing dust!! {sbnltk_default.bcolors.ENDC}")
            return word

    def dust_removal_sent(self,sentence):
        words=sentence.split()
        temp=[]
        for i in words:
            temp.append(self.dust_removal(i))
        temp=' '.join(temp)
        return  temp
    def stopword_remove(self,text):
        try:
            querywords = text.split()
            resultwords = [word for word in querywords if word not in self.__stopwords]
            result = ' '.join(resultwords)
            return result
        except:
            print(f"{sbnltk_default.bcolors.FAIL} ERROR 203: Error in Removing stop word!! {sbnltk_default.bcolors.ENDC}")
            return text

    def word_normalize(self,word):
        try:
            s = ""
            for c in word:
                g = c.encode("unicode_escape")
                g = g.upper()
                g = g[2:]
                g = g.decode('utf-8')
                if g in StaticArray.bn_norm:
                    g = StaticArray.bn_norm[g].encode().decode('utf-8')
                    s+=g
                    continue
                s+=c
            return s
        except:
            print(f"{sbnltk_default.bcolors.FAIL} ERROR 204: Error in word normalization!! {sbnltk_default.bcolors.ENDC}")
            return word


    def bangla_to_english_Conversion(self,word):
        try:
            s=""
            for c in word:
                g=c.encode("unicode_escape")
                g=g.upper()
                g=g[2:]
                g=g.decode('utf-8')
                if g in StaticArray.bn2enPunc:
                    if len(s)>0 and s[-1]=='a':
                        s=s[:-1]
                    s+=StaticArray.bn2enPunc[g]
                    continue
                if g in StaticArray.bn2en:
                    s+=StaticArray.bn2en[g]
            return s
        except:
            print(f"{sbnltk_default.bcolors.FAIL} ERROR 205: Error in Bangla to English Conversion!! {sbnltk_default.bcolors.ENDC}")
            return word


    def __bnCompare(self,item1,item2):
        g1=self.bangla_to_english_Conversion(item1)
        g2=self.bangla_to_english_Conversion(item2)
        return (g1 > g2) - (g1 < g2)

    def isBanglaWord(self,word):
        if word in self.__word_list:
            return True
        return False

    def isBangla(self,word):
        for c in word:
            g = c.encode("unicode_escape")
            g = g.upper()
            g = g[2:]
            g = g.decode('utf-8')
            if g in StaticArray.bn2en:
                return True
        return False

    def bn_word_sort_en_sys(self,vec):
        try:
            temp_vec=[]
            for i in vec:
                if self.isBangla(i):
                  i=self.dust_removal(i)
                  temp_vec.append(self.punctuation_remove(i).replace(' ',''))
            vec=list(set(temp_vec))
            vec=sorted(vec,key=functools.cmp_to_key(self.__bnCompare))
            return vec
        except:
            print(f"{sbnltk_default.bcolors.FAIL} ERROR 206: Error in Sort bangla words according English alphabet!! {sbnltk_default.bcolors.ENDC}")
            return vec

    def __bnCompare2(self,item1,item2):
        ln=min(len(item1),len(item2))
        for i in range(ln):
            if item1[i]==item2[i]:
                continue
            g1 = item1[i].encode("unicode_escape")
            g1 = g1.upper()
            g1 = g1[2:]
            g1 = g1.decode('utf-8')
            g1=StaticArray.bn_serial[g1]
            g2 = item2[i].encode("unicode_escape")
            g2 = g2.upper()
            g2 = g2[2:]
            g2 = g2.decode('utf-8')
            g2=StaticArray.bn_serial[g2]
            return (g1>g2) -(g1<g2)
        return (len(item1)>len(item2))-(len(item1)<len(item2))
    def bn_word_sort_bn_sys(self,vec):
        try:
            temp_vec = []
            for i in vec:
                if self.isBangla(i):
                    i = self.dust_removal(i)
                    temp_vec.append(self.punctuation_remove(i).replace(' ', ''))
            vec = list(set(temp_vec))
            vec = sorted(vec, key=functools.cmp_to_key(self.__bnCompare2))
            return vec
        except:
            print(f"{sbnltk_default.bcolors.FAIL} ERROR 207: Error in Sort Bangla words according Bangla alphabet!! {sbnltk_default.bcolors.ENDC}")
            return vec

    def is_number(self,word):
        for c in word:
            g = c.encode("unicode_escape")
            g = g.upper()
            g = g[2:]
            g = g.decode('utf-8')
            if g in StaticArray.bn2enNum:
                return True
        return False
    def extra_space_remove(self,sent):
        while len(sent)>0 and sent[0]==' ':
            sent=sent[1:]
        temp=''
        for i in sent:
            if len(temp)>0 and temp[-1]==' ' and i==' ':
                continue
            temp+=i
        return temp








from sbnltk.Preprocessor import preprocessor,StaticArray
from sbnltk.Tokenizer import wordTokenizer
import functools
import re
from sbnltk import sbnltk_default
from sbnltk.Downloader import downloader

rule_words = ['ই', 'ও', 'তো', 'কে', 'তে', 'রা', 'চ্ছি', 'চ্ছিল', 'চ্ছে', 'চ্ছিস', 'চ্ছিলেন', 'চ্ছ', 'য়েছে', 'েছ', 'েছে',
              'েছেন', 'রছ', 'রব', 'েল', 'েলো', 'ওয়া', 'েয়ে', 'য়', 'য়ে', 'েয়েছিল', 'েছিল', 'য়েছিল', 'েয়েছিলেন',
              'ে.েছিলেন', 'েছিলেন', 'লেন', 'দের', 'ে.ে', 'ের', 'ার', 'েন', 'বেন', 'িস', 'ছিস', 'ছিলি', 'ছি', 'ছে', 'লি',
              'বি', 'ে', 'টি', 'টির', 'েরটা', 'েরটার', 'টা', 'টার', 'গুলো', 'গুলোর', 'েরগুলো', 'েরগুলোর'
              'যোগ্য','কেই','েও','সহ','রা','ভাবে','কারি','কৃত','ই','কে','র','কি','েই','ভাবে','গুলো',
              'তে','েতে','গন','মুলক','সুচক','টুকু','টুকুই','গুলি','পদ','সমুহ','সংক্রান্ত','সংলগ্ন','সংশ্লিষ্ট',
              'সুত্রে','রুপে','ানুসারে','ানুযায়ি','তত্ত্','ি','মুখি','প্রতি','ভাবে','য়'
              ]
rule_dict = {"রছ":"র","রব":"র","েয়ে":"া","েয়েছিল":"া","েয়েছিলেন":"া","ে.েছিলেন":"া.","ে.ে":"া."}


class stemmerOP:
    __wordtokens = wordTokenizer()
    __word_vec = []
    __word_dict = {}
    __word_dict2 = {}
    __bp = preprocessor()
    __dl=downloader()
    def __init__(self):
        self.__dl.download('rootword_list',sbnltk_default.sbnltk_root_path+'dataset/')
        self.__dl.download('ner_static',sbnltk_default.sbnltk_root_path+'dataset/')
        for word in open(sbnltk_default.sbnltk_root_path + 'dataset/ner_static.txt', "r"):
            word = word.replace('\n', '')
            segment = word.split(' ')
            word = segment[:-1]
            for i in word:
                self.__word_dict[i]=1
        for word in open(sbnltk_default.sbnltk_root_path+'dataset/rootword_list.txt', "r"):
            word=word.replace('\n','')
            self.__word_dict2[word]=1
    def __search(self,word):
        if (self.__bp.word_normalize(word) in self.__word_dict) or (word in self.__word_dict) or (word in self.__word_dict2) or (self.__bp.word_normalize(word) in self.__word_dict2):
            return True
        return False
    def __bnCompare(self,item1,item2):
        return (len(item1)<len(item2))-(len(item1)>len(item2))

    def stemWord(self,word):
        try:
            if self.__word_dict2.get(word)!=None:
                return word
            suf_arr=[]
            for wd in rule_words:
                if re.search('.*' + wd + '$', word):
                    suf_arr.append(wd)
            suf_arr = sorted(suf_arr, key=functools.cmp_to_key(self.__bnCompare))
            if len(suf_arr)>0:
                for i in suf_arr:
                    if i in rule_dict:
                        ind = len(word) - len(i)
                        new_word=word[0:ind]+rule_dict[i]
                        if self.__search(new_word):
                            return new_word
                    ind = len(word) - len(i)
                    new_word = word[0:ind]
                    if len(new_word)==0:
                        return word
                    if self.__search(new_word):
                        return new_word
            return word
        except:
            print(f"{sbnltk_default.bcolors.FAIL}ERROR 101: Error in stemming!! {sbnltk_default.bcolors.ENDC}")
    def stemSent(self,sent):
        tokens=self.__wordtokens.basic_tokenizer(sent)
        temp_tokens=[]
        for i in tokens:
            temp_tokens.append(self.stemWord(i))
        result = ' '.join(temp_tokens)

        return result







import numpy as np
from sbnltk import sbnltk_default
from sbnltk.Downloader import downloader
from scipy import spatial
import  fasttext
import pickle

class fasttext_embedding:
    __dl=downloader()
    __w2v=None
    def __init__(self):
        self.__dl.download('fasttext_w2v',sbnltk_default.sbnltk_root_path+'model/')
        self.__w2v=fasttext.load_model(sbnltk_default.sbnltk_root_path+'model/fasttext_w2v.model')
    def get_vector(self,word):
        try:
            return self.__w2v[word]
        except:
            raise ValueError('Sorry!! Word is not exist in vocab!!')
    def get_nearest_neighbors(self,word,n=5):
        return self.__w2v.get_nearest_neighbors(word,k=n)

    def cosine_distance(self,word1,word2):
        if word1==word2:
            return 1.0
        try:
            vec1=self.__w2v[word1]
        except:
            raise ValueError('Sorry!! 1st word is not exist in vocab!!')
        try:
            vec2=self.__w2v[word2]
        except:
            raise ValueError('Sorry!! 2nd word is not exist in vocab!!')

        return (1.0-spatial.distance.cosine(vec1,vec2))

class gensim_word2vec_embedding:
    __dl=downloader()
    __embeddings_dict = {}
    def __init__(self):
        self.__dl.download('gensim_w2v',sbnltk_default.sbnltk_root_path+'model/')
        path=sbnltk_default.sbnltk_root_path+'model/gensim_w2v.txt'
        with open(path, 'r') as f:
            for l in f:
                values = l.split()
                word = str(values[0])
                vec = np.asarray(values[1:], "float32")
                if len(vec)<100:
                    continue
                self.__embeddings_dict[word] = vec
    def get_vector(self,word):
        if word in self.__embeddings_dict:
           return self.__embeddings_dict[word]
        return np.zeros(100)
    def cosine_distance(self,word1,word2):
        if word1==word2:
            return 1.0
        vec1=np.zeros(100)
        vec2=np.zeros(100)
        flg=0
        if word1 in self.__embeddings_dict:
            vec1=self.__embeddings_dict[word1]
            flg+=1
        if word2 in self.__embeddings_dict:
            vec2=self.__embeddings_dict[word2]
            flg+=1
        if flg<=1:
            return 0.5
        d = 1.0 - spatial.distance.cosine(vec1,vec2)
        return d
    def get_nearest_neighbors(self,item,n):
        vec = []
        if item not in self.__embeddings_dict:
            vec.append(item)
            return vec
        result=sorted(self.__embeddings_dict.keys(), key=lambda word: spatial.distance.euclidean(self.__embeddings_dict[word], self.__embeddings_dict[item]))
        j=0
        for i in result:
           if j>=n:
             break
           vec.append(i)
           j+=1

        return vec

class glove_embedding:
    __dl=downloader()
    __embeddings_dict = {}
    def __init__(self):
        self.__dl.download('glove_embedding',sbnltk_default.sbnltk_root_path+'model/')
        self.__dl.download('glove_id2word', sbnltk_default.sbnltk_root_path + 'model/')
        path=sbnltk_default.sbnltk_root_path+'model/glove_embedding.pkl'
        model=pickle.load(open(path,'rb'))
        id2word=sbnltk_default.sbnltk_root_path+'model/glove_id2word.txt'
        with open(id2word, 'r') as f:
            for l in f:
                values = l.split()
                ind = int(values[0])
                word=str(values[1])
                vec = model[ind]
                if len(vec)<100:
                    continue
                self.__embeddings_dict[word] = vec

    def get_vector(self,word):
        if word in self.__embeddings_dict:
           return self.__embeddings_dict[word]
        return np.zeros(100)
    def cosine_distance(self,word1,word2):
        vec1=np.zeros(100)
        vec2=np.zeros(100)
        flg=0
        if word1 in self.__embeddings_dict:
            vec1=self.__embeddings_dict[word1]
            flg+=1
        if word2 in self.__embeddings_dict:
            vec2=self.__embeddings_dict[word2]
            flg+=1
        if flg==0:
            return 0.5
        d = 1.0 - spatial.distance.cosine(vec1,vec2)
        return d
    def get_nearest_neighbors(self,item,n):
        vec = []
        if item not in self.__embeddings_dict:
            vec.append(item)
            return vec
        result=sorted(self.__embeddings_dict.keys(), key=lambda word: spatial.distance.euclidean(self.__embeddings_dict[word], self.__embeddings_dict[item]))
        j=0
        for i in result:
           if j>=n:
             break
           vec.append(i)
           j+=1

        return vec





import gdown
import os
from sbnltk import sbnltk_default
from os import path

class downloader():
    __root_path = sbnltk_default.sbnltk_root_path
    download_link={}            # for reserved link
    download_link_default="142XvJg9xdpgzuYD31Y4pm-ZVdMaWmtuq"    # deafault download link
    url_prefix="https://drive.google.com/uc?id="    # download url prefix


    def __init__(self):
        try:
            #dataset and model directory create
            if path.exists(self.__root_path+'dataset')==False:
                os.mkdir(self.__root_path+'dataset')
            if path.exists(self.__root_path+'model')==False:
                os.mkdir(self.__root_path+'model')

            #download link download if it is not exist
            if path.exists(self.__root_path+'dataset/download_link.txt')==False:
                url = self.url_prefix+self.download_link_default;
                output = self.__root_path+'dataset/download_link.txt'
                gdown.download(url, output, quiet=False)

            # mapping download link
            for line in open(self.__root_path+'dataset/download_link.txt','r'):
                line=line.rstrip('\n')
                items=line.split(' ')
                self.download_link[items[0]]=(items[1],items[2])    # Header = (download_id, fileType)
        except:
            raise ValueError("Error in loading download link! Check Internet connection And Try again!")

    '''
    model_path='model/'     for models
    model_path='dataset/'   for dataset
    '''

    def download(self,model,model_path):
        try:
            if self.download_link.get(model)==None:
                raise ValueError('Model Name does not exists!! ')

            model_path=model_path+model+'.'+self.download_link[model][1]
            if path.exists(model_path)==True:
                pass
            else:
                url = self.url_prefix + self.download_link[model][0]
                output = model_path
                gdown.download(url, output, quiet=False)
        except:
            raise ValueError("Error when downloading model!! Check internet Connection!!")

    def remove(self,model_path):
        try:
            os.remove(model_path)
            print(f"{sbnltk_default.bcolors.OKGREEN} {model_path} is removed!! {sbnltk_default.bcolors.ENDC} ")
        except:
            raise  ValueError("Model path does not exist!!")


    def reDownload(self,model,model_path):
        try:
            if self.download_link.get(model)==None:
                raise ValueError("Model path does not exist!!")
            model_path=model_path+model+'.'+self.download_link[model][1]
            if path.exists(model_path)==True:
                self.remove(model_path)
            url = self.url_prefix + self.download_link[model][0]
            output = model_path
            gdown.download(url, output, quiet=False)
            print(f"{sbnltk_default.bcolors.OKGREEN} Download Completed!! {sbnltk_default.bcolors.ENDC} ")
        except:
            raise ValueError("Error when downloading model!! Check Internet Connection")



import os
sbnltk_root_path = os.path.dirname(__file__)+'/'
bangla_full_stop='।'
class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'




'''
FBM
TBM
W2vBM
FBM+TBM
FBM+W2VBM
----------
with title
without title
------------
'''

from sbnltk.FBM_Information_Extraction import feature_based_information_extraction
from sbnltk.Semantic_Information_Extraction import information_extraction_word2vec
from sbnltk.Semantic_Information_Extraction import information_extraction_transformer_embedding
from sbnltk.Tokenizer import sentenceTokenizer

class feature_based_model:
    __fbm=None
    def __init__(self):
        self.__fbm=feature_based_information_extraction()

    def generate_summary(self,text,title=None,max_length=5):
        return self.__fbm.extract_text_to_text(text,title,max_length)

class word2vec_based_model:
    __w2vbm=None
    __sentT=None
    def __init__(self):
        self.__w2vbm=information_extraction_word2vec()
        self.__sentT=sentenceTokenizer()
    def generate_summary(self,text,title=None,max_length=5):
        sentences=self.__sentT.basic_tokenizer(text)
        if title==None:
            temp=self.__w2vbm.compute_information_from_single_centroid(sentences,max_length)
            sum_text = []
            for i in temp:
                sum_text.append(i[2])
            sum_text=self.__sentT.sentence_vector_to_text(sum_text)
            return sum_text
        temp= self.__w2vbm.compute_information_from_title(title,sentences,max_length)
        sum_text = []
        for i in temp:
            sum_text.append(i[2])
        sum_text=self.__sentT.sentence_vector_to_text(sum_text)
        return sum_text

class transformer_based_model:
    __tbm=None
    __sentT=None
    def __init__(self):
        self.__sentT=sentenceTokenizer()
        self.__tbm=information_extraction_transformer_embedding()
    def generate_summary(self,text,title=None,max_length=5):
        sentences=self.__sentT.basic_tokenizer(text)
        if title==None:
            temp=self.__tbm.compute_information_from_centroid(sentences,max_length)
            sum_text=[]
            for i in temp:
                sum_text.append(i[2])
            sum_text=self.__sentT.sentence_vector_to_text(sum_text)
            return sum_text
        temp= self.__tbm.compute_information_from_title(sentences,max_length)
        sum_text = []
        for i in temp:
            sum_text.append(i[2])
        sum_text=self.__sentT.sentence_vector_to_text(sum_text)
        return sum_text




'''
We used Google translator here.
google_trans_new python package.
We use 2 seconds delay for large data in google translator
We use autodetect for large translator call
'''
import time
from time import time
from sbnltk import sbnltk_default
from google_trans_new import google_translator
translator = google_translator()

class bangla_google_translator():

    def __init__(self):
        try:
            file=open(sbnltk_default.sbnltk_root_path+'dataset/auto_detect_large_translate.txt','w')
            file.write(str(str(0.0)+' '+'0'))
        except:
            print(f"{sbnltk_default.bcolors.FAIL}ERROR 101: Error in auto detecting text file initializing!! {sbnltk_default.bcolors.ENDC}")


    def translate_E2B(self,sentence):
        try:
            with open(sbnltk_default.sbnltk_root_path+'dataset/auto_detect_large_translate.txt','r') as f:
                line=f.readline()
                item=line.split(' ')
                last_epoch = float(item[0])
                epoch_count=int(item[1])
            now=float(time())
            if now-last_epoch<1.0 and epoch_count>500:
                time.sleep(2)
            if now-last_epoch>1000.0:
                epoch_count=0
            epoch_count+=1
            translation=translator.translate(sentence,lang_tgt='bn')
            now=float(time())
            with open(sbnltk_default.sbnltk_root_path+'dataset/auto_detect_large_translate.txt','w') as f:
                f.write(str(str(now)+' '+str(epoch_count)))
                f.close()
            return translation
        except:
            print(f"{sbnltk_default.bcolors.FAIL}ERROR 102: Error in E2B translation!! Check connection{sbnltk_default.bcolors.ENDC}")
            return sentence

    def translate_B2E(self,sentence):
        try:
            with open(sbnltk_default.sbnltk_root_path+'dataset/auto_detect_large_translate.txt','r') as f:
                line=f.readline()
                item=line.split(' ')
                last_epoch = float(item[0])
                epoch_count=int(item[1])
            now=float(time())
            if now-last_epoch<1.0 and epoch_count>500:
                time.sleep(2)
            if now-last_epoch>1000.0:
                epoch_count=0
            epoch_count+=1
            translation=translator.translate(sentence,lang_tgt='en')
            now=float(time())
            with open(sbnltk_default.sbnltk_root_path+'dataset/auto_detect_large_translate.txt','w') as f:
                f.write(str(str(now)+' '+str(epoch_count)))
                f.close()
            return translation
        except:
            print(f"{sbnltk_default.bcolors.FAIL}ERROR 103: Error in B2E translation!! Check connection{sbnltk_default.bcolors.ENDC}")
            return sentence




import re
from sbnltk.Ner_bert import Bert_Multilingual_Uncased_Ner
from sbnltk.Tokenizer import wordTokenizer,sentenceTokenizer

class plural:
    __suffix=['গুলো','গুলোর','েরগুলো','েরগুলোর']
    def is_plural(self,word):
        suf_arr = []
        for wd in self.__suffix:
            if re.search('.*' + wd + '$', word):
                suf_arr.append(wd)
        if len(suf_arr)>0:
            return True
        return False

class static:
    __suffix=['ই', 'ও', 'কে', 'তে', 'রা', 'য়', 'য়ে','ের', 'ার', 'েন','ে','টা', 'টার', 'গুলো', 'গুলোর', 'েরগুলো', 'েরগুলোর'
              'যোগ্য','কেই','েও','সহ','রা','ই','কে','র','কি','েই','গুলো',
              'তে','েতে','ি','য়','টি', 'টির']
    __loc=['স্থান','জায়গা','অবস্থান','স্থানটি','শহর','রাজধানী','নগর','এলাকা','অঞ্চল','বন্দর','প্রদেশ','গ্রাম','দেশ','রাষ্ট্র']
    __org=['সংগঠন','প্রতিষ্ঠান','সংস্থা','সংঘ','দল','সম্প্রদায়','এনজিও','গোষ্ঠী','সমাজ','ফোরাম']
    __per=['ব্যক্তি','মানুষ','লোক','নারী','মহিলা','ছেলে']
    __date=['তারিখ','দিন','সময়','বছর','মাস','সপ্তাহ','দিন','শতক']
    def isLoc(self,word):
        suf_arr = []
        for wd in self.__suffix:
            if re.search('.*' + wd + '$', word):
                suf_arr.append(wd)
        for i in suf_arr:
            ind = len(word) - len(i)
            new_word = word[0:ind]
            if new_word in self.__loc:
                return True
        return False
    def isOrg(self,word):
        suf_arr = []
        for wd in self.__suffix:
            if re.search('.*' + wd + '$', word):
                suf_arr.append(wd)
        for i in suf_arr:
            ind = len(word) - len(i)
            new_word = word[0:ind]
            if new_word in self.__org:
                return True
        return False
    def isPer(self,word):
        suf_arr = []
        for wd in self.__suffix:
            if re.search('.*' + wd + '$', word):
                suf_arr.append(wd)
        for i in suf_arr:
            ind = len(word) - len(i)
            new_word = word[0:ind]
            if new_word in self.__per:
                return True
        return False
    def isDate(self,word):
        suf_arr = []
        for wd in self.__suffix:
            if re.search('.*' + wd + '$', word):
                suf_arr.append(wd)
        for i in suf_arr:
            ind = len(word) - len(i)
            new_word = word[0:ind]
            if new_word in self.__date:
                return True
        return False

class Pronoun:
    __person_singular=['আমি','তুই','তুমি','আপনি','ইনি','উনি','সে','তিনি']
    __person_singular_f=['ও','এ']
    __object_singular=['এটি','এটা','ওটি','ওটা','সেটি','সেটা']
    __person_objective_singular=['আমাকে','তোকে','তোমাকে','আপনাকে','একে','এঁকে','তাকে','তাঁকে','ওঁকে','ওকে']
    __person_possessive_singular=['আমার','তোর','তোমার','আপনার','তার','তাঁর']
    __object_possessive_singular=['এটির','এটার','ওটির','ওটার','সেটির','সেটার']
    __person_plural=['আমরা','তোরা','তোমরা','আপনারা','এরা','এঁরা','ওরা','ওঁরা','তারা','তাঁরা']
    __object_plural=['এগুলো','ওগুলো','সেগুলো']
    __person_objective_plural=['আমাদেরকে','তোদেরকে','তোমাদেরকে','আপনাদেরকে','এদেরকে','এঁদেরকে','ওদেরকে','ওঁদেরকে','তাদেরকে','তাঁদেরকে']
    __object_objective_plural=['এগুলো','ওগুলো','সেগুলো']
    __person_possessive_plural=['আমাদের','তোদের','তোমাদের','আপনাদের','এদের','এঁদের','ওদের','ওঁদের','তাদের','তাঁদের']
    __object_possessive_plural=['এগুলোর','ওগুলোর','সেগুলোর']
    __own_singular=['নিজেকে','নিজে','নিজেই','নিজের','নিজেরই','নিজেরও']
    __own_plural=['নিজেদের','নিজেরা','নিজেরাই']
    __all=['এই']
    def isPronoun(self,word):
        if word in self.__person_singular:
            return (True,'PS')
        if word in self.__object_singular:
            return (True,'OS')
        if word in self.__person_objective_singular:
            return (True,'POS')
        if word in self.__person_possessive_singular:
            return (True,'PPS')
        if word in self.__object_possessive_singular:
            return (True,'OPS')
        if word in self.__person_plural:
            return (True,'PP')
        if word in self.__object_plural:
            return (True,'OP')
        if word in self.__person_objective_plural:
            return (True,'POP')
        if word in self.__object_objective_plural:
            return (True,'OOP')
        if word in self.__person_possessive_plural:
            return (True,'PPP')
        if word in self.__object_possessive_plural:
            return (True,'OPP')
        if word in self.__own_singular:
            return (True, 'OWS')
        if word in self.__own_plural:
            return (True,'OWP')
        if word in self.__all:
            return (True,'A')
        return (False,'O')

class coreference_resolution:
    __nerT=None
    __wordT=None
    __sentT=None
    __nerS=None
    __pro = Pronoun()
    __plu = plural()
    __sta = static()
    __last_person_singular = []
    __last_person_plural = []
    __last_object_singular = []
    __last_object_plural = []
    __last_loc_singular = []
    __last_date_singular = []
    __last_org_singular = []
    __last_loc_plural = []
    __last_date_plural = []
    __last_org_plural = []
    def __init__(self):
        self.__nerT=Bert_Multilingual_Uncased_Ner()
        self.__wordT=wordTokenizer()
        self.__sentT=sentenceTokenizer()
    def __cluster(self,vec):
        temp=[]
        lst=""
        ent=[]
        k=1
        for i in vec:
            for j in i:
                if(i[j][0]=='S'):
                    if len(ent)>0:
                        l=len(ent)
                        ent=' '.join(ent)
                        temp.append((k-l,k-1,ent,lst))
                        ent=[]
                        lst=""
                    temp.append((k,k,j,i[j][2:]))
                elif (i[j][0]=='B'):
                    if len(ent)>0:
                        l = len(ent)
                        ent=' '.join(ent)
                        temp.append((k-l,k-1,ent,lst))
                        ent=[]
                        lst=""
                    else:
                        ent.append(j)
                        lst=i[j][2:]
                elif (i[j][0]=='I' or i[j][0]=='E'):
                    tm=i[j][2:]
                    if tm!=lst and lst!="":
                        l = len(ent)
                        ent = ' '.join(ent)
                        temp.append((k-l, k-1,ent, lst))
                        ent = []
                        lst = ""
                    else:
                        ent.append(j)
                        lst = i[j][2:]
                else:
                    if len(ent)>0:
                        l = len(ent)
                        ent=' '.join(ent)
                        temp.append((k-l,k-1,ent,lst))
                        ent=[]
                        lst=""
                    if j=="ও":
                        temp.append((k,k,j, "CON"))
            k+=1
        if len(ent) > 0:
            l = len(ent)
            ent = ' '.join(ent)
            temp.append((k - l, k - 1, ent, lst))
            ent = []
            lst = ""
        return temp

    def __get_entity(self,sentences):
        temp2=self.__nerT.tag(sentences)
        entity=[]
        mp=[]
        mp2=[]
        for i in range(len(temp2)):
            z=self.__cluster(temp2[i])
            z=list(set(z))
            entity.append(z)
            tmp={}
            tmp2={}
            for j in range(len(z)):
                 k=z[j]
                 tmp[(int)(k[0])]=(k[2],k[3])
                 tmp2[(int)(k[1])]=(k[2],k[3])
            mp.append(tmp)
            mp2.append(tmp2)
        return mp,mp2

    def __set_ent(self,word1,ent,ind,word_i,word2=None):
        if ent=='PER':
            if word2==None:
                self.__last_person_singular.append((word1,ind,word_i))
            elif self.__plu.is_plural(word1) and word2==None:
                self.__last_person_plural.append((word1, ind,word_i))
            else:
                self.__last_person_singular.append((word1,ind,word_i))
                self.__last_person_singular.append((word2,ind,word_i))
                self.__last_person_plural.append((word1+' ও '+word2,ind,word_i))
        if ent=='OBJ':
            if word2==None:
                if self.__plu.is_plural(word1):
                    self.__last_object_plural.append((word1, ind,word_i))
                else:
                    self.__last_object_singular.append((word1,ind,word_i))
            else:
                if self.__plu.is_plural(word1)==False and self.__plu.is_plural(word2)==False:
                    self.__last_object_singular.append((word1,ind,word_i))
                    self.__last_object_singular.append((word2,ind,word_i))
                    self.__last_object_plural.append((word1 + ' ও ' + word2, ind,word_i))
                else:
                    self.__last_object_plural.append((word1, ind,word_i))
                    self.__last_object_plural.append((word2, ind,word_i))
        if ent=='LOC':
            if word2==None:
                self.__last_loc_singular.append((word1,ind,word_i))
            elif self.__plu.is_plural(word1) and word2==None:
                self.__last_loc_plural.append((word1, ind,word_i))
            else:
                self.__last_loc_singular.append((word1,ind,word_i))
                self.__last_loc_singular.append((word2,ind,word_i))
                self.__last_loc_plural.append((word1+' ও '+word2,ind,word_i))
        if ent=='ORG':
            if word2==None:
                self.__last_org_singular.append((word1,ind,word_i))
            elif self.__plu.is_plural(word1) and word2==None:
                self.__last_org_plural.append((word1, ind,word_i))
            else:
                self.__last_org_singular.append((word1,ind,word_i))
                self.__last_org_singular.append((word2,ind,word_i))
                self.__last_org_plural.append((word1+' ও '+word2,ind,word_i))
        if ent=='DATE':
            if word2==None:
                self.__last_date_singular.append((word1,ind,word_i))
            elif self.__plu.is_plural(word1) and word2==None:
                self.__last_date_plural.append((word1, ind,word_i))
            else:
                self.__last_date_singular.append((word1,ind,word_i))
                self.__last_date_singular.append((word2,ind,word_i))
                self.__last_date_plural.append((word1+' ও '+word2,ind,word_i))

    # For renew the entity, After resolving i push it again in entity list
    def __set_ent_renew(self,word,type,i,word_i):
        if (type=='PS'):
            self.__last_person_singular.append((word,i,word_i))
        elif type=='PP':
            self.__last_person_plural.append((word, i,word_i))
        elif type=='OS':
            self.__last_object_singular.append((word, i,word_i))
        elif type=='OP':
            self.__last_object_plural.append((word, i,word_i))
        elif type=='LOCS':
            self.__last_loc_singular.append((word, i,word_i))
        elif type=='LOCP':
            self.__last_loc_plural.append((word, i,word_i))
        elif type=='ORGS':
            self.__last_org_singular.append((word, i,word_i))
        elif type=='ORGP':
            self.__last_org_plural.append((word, i,word_i))
        elif type=='DATES':
            self.__last_date_singular.append((word, i,word_i))
        elif type=='DATEP':
            self.__last_date_plural.append((word, i,word_i))
    def __is_last_ch_vowel(self,word):
        tag=['া','ি','ী','ু','ূ','ৃ','ে','ৈ','ো','ৌ','আ','ই','ঈ','উ','ঊ','এ','ঐ','ও','ঔ']
        if len(word)>0 and word[-1] in tag:
            return True
        return False


    # main resolve program
    def __single_resolve(self,type,i,word,word_i):
        if type == 'PS' and len(self.__last_person_singular) > 0:
            ind = len(self.__last_person_singular) - 1
            flg = -1
            flg2=-1
            temp_ref = ""
            while (ind >= 0):
                if flg == -1:
                    flg = self.__last_person_singular[ind][1]
                    flg2= self.__last_person_singular[ind][2]
                    temp_ref = self.__last_person_singular[ind][0]
                elif flg != -1 and self.__last_person_singular[ind][1] == flg:
                    flg = self.__last_person_singular[ind][1]
                    flg2 = self.__last_person_singular[ind][2]
                    temp_ref = self.__last_person_singular[ind][0]
                ind -= 1
            if flg == -1:
                self.__set_ent_renew(self.__last_person_singular[-1][0],'PS',i,word_i)
                return [self.__last_person_singular[-2][1]],[self.__last_person_singular[-2][2]],self.__last_person_singular[-1][0]
            else:
                self.__set_ent_renew(temp_ref, 'PS', i,word_i)
                return [flg],[flg2],temp_ref
        elif type == 'POS' and len(self.__last_person_singular) > 0:
            ind = len(self.__last_person_singular) - 1
            flg = -1
            flg2=-1
            temp_ref = ""
            while (ind >= 0):
                if flg == -1 and self.__last_person_singular[ind][1] != i:
                    flg = self.__last_person_singular[ind][1]
                    flg2=self.__last_person_singular[ind][2]
                    temp_ref = self.__last_person_singular[ind][0]
                ind -= 1
            if flg == -1:
                wd = self.__last_person_singular[-1][0]
                self.__set_ent_renew(wd, 'PS', i,word_i)
                if len(wd) >= 2 and wd[len(wd) - 2:] == 'কে':
                    pass
                else:
                    wd += 'কে'
                return [self.__last_person_singular[-2][1]],[self.__last_person_singular[-2][2]],wd
            else:
                wd = temp_ref
                if len(wd) >= 2 and wd[len(wd) - 2:] == 'কে':
                    pass
                else:
                    wd += 'কে'
                self.__set_ent_renew(temp_ref, 'PS', i,word_i)
                return [flg],[flg2],wd
        elif type == 'PPS' and len(self.__last_person_singular) > 0:
            ind = len(self.__last_person_singular) - 1
            flg = -1
            flg2=-1
            temp_ref = ""
            while (ind >= 0):
                if flg == -1:
                    flg = self.__last_person_singular[ind][1]
                    flg2=self.__last_person_singular[ind][2]
                    temp_ref = self.__last_person_singular[ind][0]
                elif flg != -1 and self.__last_person_singular[ind][1] == flg:
                    flg = self.__last_person_singular[ind][1]
                    flg2 = self.__last_person_singular[ind][2]
                    temp_ref = self.__last_person_singular[ind][0]
                ind -= 1
            if flg == -1:
                wd = self.__last_person_singular[-1][0]
                self.__set_ent_renew(wd, 'PS', i,word_i)
                if self.__is_last_ch_vowel(wd):
                    wd += 'র'
                else:
                    wd += 'ের'
                return [self.__last_person_singular[-2][1]],[self.__last_person_singular[-2][2]],wd
            else:
                wd = temp_ref
                if self.__is_last_ch_vowel(wd):
                    wd += 'র'
                else:
                    wd += 'ের'
                self.__set_ent_renew(temp_ref, 'PS', i,word_i)
                return [flg],[flg2],wd



        elif type == 'OS' and len(self.__last_object_singular) > 0:
            ind = len(self.__last_object_singular) - 1
            flg = -1
            flg2=-1
            temp_ref = ""
            while (ind >= 0):
                if flg == -1:
                    flg = self.__last_object_singular[ind][1]
                    flg2=self.__last_object_singular[ind][2]
                    temp_ref = self.__last_object_singular[ind][0]
                elif flg != -1 and self.__last_object_singular[ind][1] == flg:
                    flg = self.__last_object_singular[ind][1]
                    flg2 = self.__last_object_singular[ind][2]
                    temp_ref = self.__last_object_singular[ind][0]
                ind -= 1
            if flg == -1:
                wd = self.__last_object_singular[-1][0]
                self.__set_ent_renew(wd, 'OS', i,word_i)
                return [self.__last_object_singular[-2][1]],[self.__last_object_singular[-2][2]],wd
            else:
                wd = temp_ref
                self.__set_ent_renew(temp_ref, 'OS', i,word_i)
                return [flg],[flg2],wd
        elif type == 'OOS' and len(self.__last_object_singular) > 0:
            ind = len(self.__last_object_singular) - 1
            flg = -1
            flg2=-1
            temp_ref = ""
            while (ind >= 0):
                if flg == -1 and self.__last_object_singular[ind][1] != i:
                    flg = self.__last_object_singular[ind][1]
                    flg2=self.__last_object_singular[ind][2]
                    temp_ref = self.__last_object_singular[ind][0]
                ind -= 1
            if flg == -1:
                wd = self.__last_object_singular[-1][0]
                self.__set_ent_renew(wd, 'OS', i,word_i)
                if len(wd) >= 2 and wd[len(wd) - 2:] == 'কে':
                    pass
                else:
                    wd += 'কে'
                return [self.__last_object_singular[-2][1]],[self.__last_object_singular[-2][2]],wd
            else:
                wd = temp_ref
                if len(wd) >= 2 and wd[len(wd) - 2:] == 'কে':
                    pass
                else:
                    wd += 'কে'
                self.__set_ent_renew(temp_ref, 'OS', i,word_i)
                return [flg],[flg2],wd
        elif type == 'OPS' and len(self.__last_object_singular) > 0:
            ind = len(self.__last_object_singular) - 1
            flg = -1
            flg2=-1
            temp_ref = ""
            while (ind >= 0):
                if flg == -1:
                    flg = self.__last_object_singular[ind][1]
                    flg2= self.__last_object_singular[ind][2]
                    temp_ref = self.__last_object_singular[ind][0]
                elif flg != -1 and self.__last_object_singular[ind][1] == flg:
                    flg = self.__last_object_singular[ind][1]
                    flg2 = self.__last_object_singular[ind][2]
                    temp_ref = self.__last_object_singular[ind][0]
                ind -= 1
            if flg == -1:
                wd = self.__last_object_singular[-1][0]
                self.__set_ent_renew(wd, 'OS', i,word_i)
                if self.__is_last_ch_vowel(wd):
                    wd += 'র'
                else:
                    wd += 'ের'

                return [self.__last_object_singular[-2][1]],[self.__last_object_singular[-2][2]],wd
            else:
                wd = temp_ref
                if self.__is_last_ch_vowel(wd):
                    wd += 'র'
                else:
                    wd += 'ের'
                self.__set_ent_renew(temp_ref, 'OS', i,word_i)
                return [flg],[flg2],wd
        elif (type == 'PP' or type == 'POP' or type == 'PPP') and (
                len(self.__last_person_plural) > 0 or len(self.__last_person_singular) >= 2):
            ind = len(self.__last_person_plural) - 1
            flg = -1
            flg2=-1
            temp_ref = ""
            while (ind >= 0):
                if flg == -1 and self.__last_person_plural[ind][1] != i:
                    flg = self.__last_person_plural[ind][1]
                    flg2= self.__last_person_plural[ind][2]
                    temp_ref = self.__last_person_plural[ind][0]
                elif flg != -1 and self.__last_person_plural[ind][1] == flg:
                    flg = self.__last_person_plural[ind][1]
                    flg2 = self.__last_person_plural[ind][2]
                    temp_ref = self.__last_person_plural[ind][0]
                ind -= 1
            if flg == -1:
                if (len(self.__last_person_singular) >= 2):
                    wd = (self.__last_person_singular[-2][0] + ' ও ' + self.__last_person_singular[-1][0])
                    self.__set_ent_renew(wd, 'PP', i,word_i)
                    return [self.__last_person_singular[-2][1],self.__last_person_singular[-1][1]],[self.__last_person_singular[-2][2],self.__last_person_singular[-1][2]],wd
                else:
                    return [-1],[-1],word
            else:
                if len(self.__last_person_singular) >= 2 and flg < self.__last_person_singular[-2][1]:

                    wd=(self.__last_person_singular[-2][0] + ' ও ' + self.__last_person_singular[-1][0])
                    self.__set_ent_renew(wd, 'PP', i,word_i)
                    return [self.__last_person_singular[-2][1],self.__last_person_singular[-1][1]],[self.__last_person_singular[-2][2],self.__last_person_singular[-1][2]],wd
                else:
                    self.__set_ent_renew(temp_ref, 'PP', i,word_i)
                    return [flg],[flg2],temp_ref
        elif (type == 'OP' or type == 'OOP' or type == 'OPP') and (
                len(self.__last_object_plural) > 0 or len(self.__last_object_singular) >= 2):
            ind = len(self.__last_object_plural) - 1
            flg = -1
            flg2=-1
            temp_ref = ""
            while (ind >= 0):
                ind -= 1
                if flg == -1 and self.__last_object_plural[ind][1] != i:
                    flg = self.__last_object_plural[ind][1]
                    flg2= self.__last_object_plural[ind][2]
                    temp_ref = self.__last_object_plural[ind][0]
                elif flg != -1 and self.__last_object_plural[ind][1] == flg:
                    flg2 = self.__last_object_plural[ind][2]
                    flg = self.__last_object_plural[ind][1]
                    temp_ref = self.__last_object_plural[ind][0]
            if flg == -1:
                if (len(self.__last_object_singular) >= 2):
                    wd=((self.__last_object_singular[-2][0] + ' ও ' + self.__last_object_singular[-1][0]))
                    self.__set_ent_renew(wd, 'OP', i,word_i)
                    return [self.__last_object_singular[-2][1],self.__last_object_singular[-1][1]],[self.__last_object_singular[-2][2],self.__last_object_singular[-1][2]],wd
                else:
                    return [-1],[-1],word
            else:
                if len(self.__last_object_singular) >= 2 and flg < self.__last_object_singular[-2][1]:
                    wd=((self.__last_object_singular[-2][0] + ' ও ' + self.__last_object_singular[-1][0]))
                    self.__set_ent_renew(wd, 'OP', i,word_i)
                    return [self.__last_object_singular[-2][1], self.__last_object_singular[-1][1]],[self.__last_object_singular[-2][2], self.__last_object_singular[-1][2]] ,wd
                else:
                    self.__set_ent_renew(temp_ref, 'OP', i,word_i)
                    return [flg],[flg2],temp_ref
        elif type=='LOCS' and len(self.__last_loc_singular)>0:
            ind = len(self.__last_loc_singular) - 1
            flg = -1
            flg2=-1
            temp_ref = ""
            while (ind >= 0):
                if flg == -1 and self.__last_loc_singular[ind][1] != i:
                    flg = self.__last_loc_singular[ind][1]
                    flg2= self.__last_loc_singular[ind][2]
                    temp_ref = self.__last_loc_singular[ind][0]
                elif flg != -1 and self.__last_loc_singular[ind][1] == flg:
                    flg = self.__last_loc_singular[ind][1]
                    flg2 = self.__last_loc_singular[ind][2]
                    temp_ref = self.__last_loc_singular[ind][0]
                ind -= 1
            if flg == -1:
                self.__set_ent_renew(self.__last_loc_singular[-1][0], 'LOCS', i,word_i)
                return [self.__last_loc_singular[-2][1]],[self.__last_loc_singular[-2][2]],self.__last_loc_singular[-1][0]
            else:
                self.__set_ent_renew(temp_ref, 'LOCS', i,word_i)
                return [flg],[flg2],temp_ref
        elif type=='LOCP' and (len(self.__last_loc_plural) > 0 or len(self.__last_loc_singular) >= 2):
            ind = len(self.__last_loc_plural) - 1
            flg = -1
            flg2=-1
            temp_ref = ""
            while (ind >= 0):
                if flg == -1 and self.__last_loc_plural[ind][1] != i:
                    flg = self.__last_loc_plural[ind][1]
                    flg2= self.__last_loc_plural[ind][2]
                    temp_ref = self.__last_loc_plural[ind][0]
                elif flg != -1 and self.__last_loc_plural[ind][1] == flg:
                    flg = self.__last_loc_plural[ind][1]
                    flg2= self.__last_loc_plural[ind][2]
                    temp_ref = self.__last_loc_plural[ind][0]
                ind -= 1
            if flg == -1:
                if (len(self.__last_loc_singular) >= 2):
                    wd = ((self.__last_loc_singular[-2][0] + ' ও ' + self.__last_loc_singular[-1][0]))
                    self.__set_ent_renew(wd, 'LOCP', i,word_i)
                    return [self.__last_loc_singular[-2][1],self.__last_loc_singular[-1][1]],[self.__last_loc_singular[-2][2],self.__last_loc_singular[-1][2]],wd
                else:
                    return [-1],[-1],word
            else:
                if len(self.__last_loc_singular) >= 2 and flg < self.__last_loc_singular[-2][1]:
                    wd=((self.__last_loc_singular[-2][0] + ' ও ' + self.__last_loc_singular[-1][0]))
                    self.__set_ent_renew(wd, 'LOCP', i,word_i)
                    return [self.__last_loc_singular[-2][1],self.__last_loc_singular[-1][1]],[self.__last_loc_singular[-2][2],self.__last_loc_singular[-1][2]],wd
                else:
                    self.__set_ent_renew(temp_ref, 'LOCP', i,word_i)
                    return [flg],[flg2],temp_ref
        elif type == 'ORGS' and len(self.__last_org_singular) > 0:
            ind = len(self.__last_org_singular) - 1
            flg = -1
            flg2=-1
            temp_ref = ""
            while (ind >= 0):
                if flg == -1 and self.__last_org_singular[ind][1] != i:
                    flg = self.__last_org_singular[ind][1]
                    flg2= self.__last_org_singular[ind][2]
                    temp_ref = self.__last_org_singular[ind][0]
                elif flg != -1 and self.__last_org_singular[ind][1] == flg:
                    flg = self.__last_org_singular[ind][1]
                    flg2 = self.__last_org_singular[ind][2]
                    temp_ref = self.__last_org_singular[ind][0]
                ind -= 1
            if flg == -1:
                self.__set_ent_renew(self.__last_org_singular[-1][0], 'ORGS', i,word_i)
                return [self.__last_org_singular[-2][1]],[self.__last_org_singular[-2][2]],self.__last_org_singular[-1][0]
            else:
                self.__set_ent_renew(temp_ref, 'ORGS', i,word_i)
                return [flg],[flg2],temp_ref
        elif type == 'ORGP' and (len(self.__last_org_plural) > 0 or len(self.__last_org_singular) >= 2):
            ind = len(self.__last_org_plural) - 1
            flg = -1
            flg2=-1
            temp_ref = ""
            while (ind >= 0):
                if flg == -1 and self.__last_org_plural[ind][1] != i:
                    flg = self.__last_org_plural[ind][1]
                    flg2= self.__last_org_plural[ind][2]
                    temp_ref = self.__last_org_plural[ind][0]
                elif flg != -1 and self.__last_org_plural[ind][1] == flg:
                    flg = self.__last_org_plural[ind][1]
                    flg2= self.__last_org_plural[ind][2]
                    temp_ref = self.__last_org_plural[ind][0]
                ind -= 1
            if flg == -1:
                if (len(self.__last_org_singular) >= 2):
                    wd = ((self.__last_org_singular[-2][0] + ' ও ' + self.__last_org_singular[-1][0]))
                    self.__set_ent_renew(wd, 'ORGP', i,word_i)
                    return [self.__last_org_singular[-2][1],self.__last_org_singular[-1][1]],[self.__last_org_singular[-2][2],self.__last_org_singular[-1][2]],wd
                else:
                    return [-1],[-1],word
            else:
                if len(self.__last_org_singular) >= 2 and flg < self.__last_org_singular[-2][1]:
                    wd=((self.__last_org_singular[-2][0] + ' ও ' + self.__last_org_singular[-1][0]))
                    self.__set_ent_renew(wd, 'ORGP', i,word_i)
                    return [self.__last_org_singular[-2][1],self.__last_org_singular[-1][1]],[self.__last_org_singular[-2][2],self.__last_org_singular[-1][2]],wd
                else:
                    self.__set_ent_renew(temp_ref, 'ORGP', i,word_i)
                    return [flg],[flg2],temp_ref
        elif type == 'DATES' and len(self.__last_date_singular) > 0:
            ind = len(self.__last_date_singular) - 1
            flg = -1
            flg2=-1
            temp_ref = ""
            while (ind >= 0):
                if flg == -1 and self.__last_date_singular[ind][1] != i:
                    flg = self.__last_date_singular[ind][1]
                    flg2= self.__last_date_singular[ind][2]
                    temp_ref = self.__last_date_singular[ind][0]
                elif flg != -1 and self.__last_date_singular[ind][1] == flg:
                    flg = self.__last_date_singular[ind][1]
                    flg2= self.__last_date_singular[ind][2]
                    temp_ref = self.__last_date_singular[ind][0]
                ind -= 1
            if flg == -1:
                self.__set_ent_renew(self.__last_date_singular[-1][0], 'DATES', i,word_i)
                return [self.__last_date_singular[-2][1]],[self.__last_date_singular[-2][2]],self.__last_date_singular[-1][0]
            else:
                self.__set_ent_renew(temp_ref, 'DATES', i,word_i)
                return [flg],[flg2],temp_ref
        elif type == 'DATEP' and (len(self.__last_date_plural) > 0 or len(self.__last_date_singular) >= 2):
            ind = len(self.__last_date_plural) - 1
            flg = -1
            flg2=-1
            temp_ref = ""
            while (ind >= 0):
                if flg == -1 and self.__last_date_plural[ind][1] != i:
                    flg = self.__last_date_plural[ind][1]
                    flg2= self.__last_date_plural[ind][2]
                    temp_ref = self.__last_date_plural[ind][0]
                elif flg != -1 and self.__last_date_plural[ind][1] == flg:
                    flg = self.__last_date_plural[ind][1]
                    flg2= self.__last_date_plural[ind][2]
                    temp_ref = self.__last_date_plural[ind][0]
                ind -= 1
            if flg == -1:
                if (len(self.__last_date_singular) >= 2):
                    wd=((self.__last_date_singular[-2][0] + ' ও ' + self.__last_date_singular[-1][0]))
                    self.__set_ent_renew(wd, 'DATEP', i)
                    return [self.__last_date_singular[-2][1],self.__last_date_singular[-1][1]],[self.__last_date_singular[-2][2],self.__last_date_singular[-1][2]],wd
                else:
                    return [-1],[-1],(word)
            else:
                if len(self.__last_date_singular) >= 2 and flg < self.__last_date_singular[-2][1]:
                    wd=((self.__last_date_singular[-2][0] + ' ও ' + self.__last_date_singular[-1][0]))
                    self.__set_ent_renew(wd, 'DATEP', i,word_i)
                    return [self.__last_date_singular[-2][1],self.__last_date_singular[-1][1]],[self.__last_date_singular[-2][2],self.__last_date_singular[-1][2]],wd
                else:
                    self.__set_ent_renew(temp_ref, 'DATEP', i,word_i)
                    return [flg],[flg2],temp_ref
        elif type=='OWS' and len(self.__last_person_singular)>0:
            if self.__last_person_singular[-1][1]==i:
                return [self.__last_person_singular[-1][1]],[self.__last_person_singular[-1][2]],self.__last_person_singular[-1][0]
            return [-1],[-1],word
        elif type=='OWP' :
            if len(self.__last_person_plural)>0 and self.__last_person_plural[-1][1]==i:
                return [self.__last_person_plural[-1][1]],[self.__last_person_plural[-1][2]],self.__last_person_plural[-1][0]
            elif len(self.__last_person_singular)>=2 and self.__last_person_singular[-1][1]==i and self.__last_person_singular[-2][1]==i:
                return [self.__last_person_singular[-2][1],self.__last_person_singular[-1][1]],[self.__last_person_singular[-2][2],self.__last_person_singular[-1][2]],(self.__last_person_singular[-2][0]+' ও '+self.__last_person_singular[-1][0])
            return [-1],[-1],word
        return [-1],[-1],word


    # connector barano lagbe
    def __tokenize(self,text):
        sentences=self.__sentT.basic_tokenizer(text)
        seg_toks=['এবং','বলেছিলেন','']
        temp=[]
        for sent in sentences:
            if ' এবং ' in sent:
                sent2=sent.split('এবং')
                for i in range(len(sent2)):
                    if i==len(sent2)-1:
                        temp.append(sent2[i])
                        break
                    temp.append(sent2[i]+'এবং')
            else:
                temp.append(sent)
        return temp



    def resolve(self,text):
        sentences=self.__tokenize(text)
        mp,mp2=self.__get_entity(sentences)
        i=0
        self.__last_person_singular = []
        self.__last_person_plural = []
        self.__last_object_singular = []
        self.__last_object_plural = []
        self.__last_date_singular = []
        self.__last_org_singular = []
        self.__last_loc_plural = []
        self.__last_date_plural = []
        self.__last_org_plural = []
        resolve_text=[]
        resolve_pos=[]
        for sent in sentences:
            words=self.__wordT.basic_tokenizer(sent)
            map2=mp2[i]
            map1=mp[i]
            i+=1
            j=0
            ref=[]
            for x in range(len(words)):
                j+=1
                if map1.get(j) and self.__pro.isPronoun(words[x])[0]==False:
                    if words[x]=='ও' and map1.get(j+1) and map2.get(j-1):
                        self.__set_ent(map2[j-1][0],map2[j-1][1],i,j,map1[j+1][0])
                    else:
                        self.__set_ent(map1[j][0], map1[j][1], i,j)
                    ref.append(([],words[x]))
                elif self.__pro.isPronoun(words[x])[0]==True:
                    type=self.__pro.isPronoun(words[x])[1]
                    if type!='A':
                        ref.append(self.__single_resolve(type,i,words[x],j))
                    elif (type=='A') and j<len(words):
                        if self.__sta.isPer(words[j]) and self.__plu.is_plural(words[j]):
                            ref.append(self.__single_resolve('PP',i,words[x],j))
                        elif self.__sta.isPer(words[j]):
                            ref.append(self.__single_resolve('PS', i, words[x],j))
                        elif self.__sta.isLoc(words[j]) and self.__plu.is_plural(words[j]):
                            ref.append(self.__single_resolve('LOCP',i,words[x],j))
                        elif self.__sta.isLoc(words[j]):
                            ref.append(self.__single_resolve('LOCS', i, words[x],j))
                        elif self.__sta.isOrg(words[j]) and self.__plu.is_plural(words[j]):
                            ref.append(self.__single_resolve('ORGP',i,words[x],j))
                        elif self.__sta.isOrg(words[j]):
                            ref.append(self.__single_resolve('ORGS', i, words[x],j))
                        elif self.__sta.isDate(words[j]) and self.__plu.is_plural(words[j]):
                            ref.append(self.__single_resolve('DATEP',i,words[x],j))
                        elif self.__sta.isDate(words[j]):
                            ref.append(self.__single_resolve('DATES', i, words[x],j))
                        else:
                            ref.append(([],words[x]))
                else:
                     ref.append(([],words[x]))
            resolve_text.append(ref)
        return resolve_text















import torch
import numpy as np
from torch.utils.data import TensorDataset, DataLoader, SequentialSampler
from keras.preprocessing.sequence import pad_sequences
from pytorch_pretrained_bert import BertTokenizer
from sbnltk.Downloader import downloader
from sbnltk import  sbnltk_default
import zipfile
import os

class bert_multilingual_cased_postag:
    __dl = downloader()
    __model = None
    __tokenizer = None
    __device = None
    __tag2idx = {'CC': 10,'CD': 8, 'DT': 6, 'IN': 5, 'JJ': 0, 'NN': 4, 'NNP': 3,'NNS': 1, 'PRE': 12, 'PRF': 9, 'PRP': 13, 'RB': 7, 'VB': 2, 'WH': 11}
    __tags2vals = {}

    # isinstance
    def __init__(self):
        self.__device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.__dl.download('bert_multi_cased_postag', sbnltk_default.sbnltk_root_path + 'model/')
        self.__dl.download('bert_vocab_postag', sbnltk_default.sbnltk_root_path + 'model/')
        self.__tokenizer = BertTokenizer.from_pretrained(
            sbnltk_default.sbnltk_root_path + 'model/bert_vocab_postag.txt')
        self.__model = torch.load(sbnltk_default.sbnltk_root_path + 'model/bert_multi_cased_postag.pth',
                                  map_location=self.__device)
        for i in self.__tag2idx:
            self.__tags2vals[self.__tag2idx[i]] = i
        self.__model.eval()

    def tag(self, sentences):

        max_seq_len = 128  # tokens
        batch_s = 8
        all_sentence_tags = []
        for sentence in sentences:
            sentence = [sentence]
            words = sentence[0].split()
            false_labels = []
            for w in range(len(words)):
                false_labels.append('NN')
            labels = [false_labels]
            tokenized_texts = [self.__tokenizer.tokenize(sent) for sent in sentence]
            X = pad_sequences([self.__tokenizer.convert_tokens_to_ids(txt) for txt in tokenized_texts],
                              maxlen=max_seq_len, dtype="long", truncating="post", padding="post")
            Y = pad_sequences([[self.__tag2idx.get(l) for l in lab] for lab in labels],
                              maxlen=max_seq_len, value=self.__tag2idx["NN"], padding="post",
                              dtype="long", truncating="post")
            attention_masks = [[float(i > 0) for i in ii] for ii in X]
            X_train = torch.tensor(X)
            Y_train = torch.tensor(Y)
            Mask_train = torch.tensor(attention_masks)
            data_valid = TensorDataset(X_train, Mask_train, Y_train)
            data_valid_sampler = SequentialSampler(data_valid)
            DL_valid = DataLoader(data_valid, sampler=data_valid_sampler, batch_size=batch_s)
            predictions = []
            for batch in DL_valid:
                batch = tuple(t.to(self.__device) for t in batch)
                b_input_ids, b_input_mask, b_labels = batch
                with torch.no_grad():
                    logits = self.__model(b_input_ids, token_type_ids=None,
                                          attention_mask=b_input_mask)
                logits = logits.detach().cpu().numpy()
                predictions.extend([list(p) for p in np.argmax(logits, axis=2)])
            pred_tags = [[self.__tags2vals[p_i] for p_i in p] for p in predictions]
            pred_tags = pred_tags[0][:(len(words))]
            temp_dict = []
            for i in range(len(words)):
                temp_dict.append((words[i], pred_tags[i]))
            all_sentence_tags.append(temp_dict)
        return all_sentence_tags

class bert_Multilingual_Uncased_Postag:
    __model = None
    __dl=downloader()
    __device=True if torch.cuda.is_available() else False
    __module_found=1
    try:
        import simpletransformers.ner.ner_model as nermodel
        __module_found=1
    except:
        __module_found=0
    def __init__(self):
        if self.__module_found==0:
            raise ValueError('Please install simpletransformers!! install Command: pip3 install simpletransformers')
        if os.path.exists(sbnltk_default.sbnltk_root_path+'model/bert_multi_uncased_postag')==False:
            self.__dl.download('bert_multi_uncased_postag', sbnltk_default.sbnltk_root_path + 'model/')
            with zipfile.ZipFile(sbnltk_default.sbnltk_root_path+'model/bert_multi_uncased_postag.zip', 'r') as file:
                file.extractall(sbnltk_default.sbnltk_root_path+'model/')
            os.remove(sbnltk_default.sbnltk_root_path+'model/bert_multi_uncased_postag.zip')
        t_h=sbnltk_default.sbnltk_root_path+'model/bert_multi_uncased_postag/model_args.json'
        t_g=sbnltk_default.sbnltk_root_path+'model/bert_multi_uncased_postag/'
        self.__model = self.nermodel.NERModel('bert',t_g, use_cuda=self.__device, args=t_h)
    def tag(self,sentences):
        d, f = self.__model.predict(sentences)
        return d



from sbnltk.Stemmer import stemmerOP
from sbnltk.Preprocessor import preprocessor
from sbnltk.Downloader import downloader
import pickle
from sbnltk import  sbnltk_default

class static_NER:
    __ner_static_data={}
    __bp = preprocessor()
    __stemmer = stemmerOP()
    __dl = downloader()
    def __init__(self):
        self.__dl.download('ner_static',sbnltk_default.sbnltk_root_path+'dataset/')
        for word in open(sbnltk_default.sbnltk_root_path+'dataset/ner_static.txt', "r"):
            word=word.replace('\n','')
            segment=word.split(' ')
            tag=segment[-1]
            word=segment[:-1]
            word=' '.join(word)

            self.__ner_static_data[word]=tag
    def tag(self,sentence):
        segment=sentence.split()
        stems=self.__stemmer.stemSent(sentence)
        stems=stems.split()
        i=0
        sentence_tags=[]
        while(i<len(segment)):
            j=len(segment)
            flg=0
            while(j>i):
                now=' '.join(segment[i:j])
                now2=' '.join(stems[i:j])
                if self.__ner_static_data.get(now)!=None:
                    sentence_tags.append((now,self.__ner_static_data[now]))
                    i=j-1
                    flg=1
                    break
                if self.__ner_static_data.get(now2)!=None:
                    sentence_tags.append((now, self.__ner_static_data[now2]))
                    i=j-1
                    flg=1
                j-=1
            if flg==0:
                sentence_tags.append((segment[i],'O'))
            i+=1
        return sentence_tags


class sklearn_NER:
    __dl=downloader()
    __bp=preprocessor()
    __sk_model=None
    def __init__(self):
        self.__dl.download('sklearn_ner',sbnltk_default.sbnltk_root_path+'model/')
        self.__sk_model = pickle.load(open(sbnltk_default.sbnltk_root_path+'model/sklearn_ner.pkl', 'rb'))
    def word2features(self,sent, i):
        return {
            'word': sent[i], 'is_first': i == 0, 'is_last': i == len(sent) - 1,
            'is_capitalized': sent[i][0].upper() == sent[i][0], 'is_all_caps': sent[i].upper() == sent[i],
            'is_all_lower': sent[i].lower() == sent[i],
            'prefix-1': sent[i][0], 'prefix-2': sent[i][:2], 'prefix-3': sent[i][:3],
            'suffix-1': sent[i][-1], 'suffix-2': sent[i][-2:], 'suffix-3': sent[i][-3:],
            'prev_word': '' if i == 0 else sent[i - 1], 'next_word': '' if i == len(sent) - 1 else sent[i + 1],
            'is_numeric': sent[i].isdigit()
        }
    def tag(self,text):
        if len(text)==0:
            return []
        words=text.split()
        sentence_features = [self.word2features(words, i) for i in range(len(words))]
        return list(zip(words, self.__sk_model.predict([sentence_features])[0]))




from Bangla_translator import bangla_google_translator
trans=bangla_google_translator()
print(trans.translate_B2E('কারাগারে থাকা ১৪ জনের মধ্যে ১২ জন স্বীকারোক্তিমূলক জবানবন্দী দিয়েছে'))
# 12 of the 14 people in prison gave confessional statements
print(trans.translate_E2B('12 of the 14 people in prison gave confessional statements'))
# কারাগারে 14 জনের মধ্যে 1২ জন স্বীকারোক্তিমূলক বক্তব্য দিয়েছেন


import pathlib
import setuptools
HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setuptools.setup(
    name="sbnltk",
    version="1.0.7",
    description="Bangla NLP toolkit",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/Foysal87/sbnltk",
    author="Towhid Ahmed Foysal",
    author_email="towhidfoysal123@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=[
'gdown>=3.12.2',
'google_trans_new>=1.1.9',
'pandas>=1.2.2',
'scikit-learn==0.22.2.post1',
'transformers>=4.3.2',
'torch',
'tensorflow>=2.4.1',
'sklearn_crfsuite>=0.3.6',
'pytorch_pretrained_bert>=0.6.2',
'sentence_transformers'
    ],
)
